/*

XNormal remote control for 3ds Max

Created:	2014-06-01
Author:		Andreas Hackel
Web page:	http://www.andreashackel.de

Version:	1.0

*/

global xNormalRolloutOpen = false;

Rollout XNormalRollout "XNormal" width:160;
(
	--label lblLowPoly "" align:#left;
	--label lblHighPoly "" align:#left;
	dropdownlist ddlSize "" items:#("128", "256", "512", "1024", "2048", "4096") selection:2 tooltip:"Bitmap Size";
	dropdownlist ddlFormat "" items:#("tga", "tif", "jpg", "png", "bmp") selection:1 tooltip:"Format";
	dropdownlist ddlAntiAliasing "" items:#("x1", "x2", "x4") selection:2 tooltip:"Antialiasing";
	dropdownlist ddlPadding "" items:#("0", "2", "4", "8", "16", "32") selection:5 tooltip:"Padding";
	label lblSpacer1 "";
	checkbox chkBaseColor "Color Map" checked:false;
	checkbox chkNormal "Normal Map" checked:true;
	checkbox chkOSNormal "Object Space Normal Map";
	checkbox chkAO "Ambient Occlusion Map";
	checkbox chkCavity "Cavity Map";
	checkbox chkConvexity "Convexity Map";
	checkbox chkVC "Bake Vertex Colors";
	checkbox chkThickness "Thickness Map";
	button btnAssignMaps "Assign maps" width:140 tooltip:"Assign selected maps to low poly mesh";
	label lblSpacer2 "";
	checkbox chkUseCage "Use Cage" checked: true tooltip:"Use a cage instead of a fixed ray distance";
	checkbox chkDeleteExportedModels "Delete Exported Models" checked: true tooltip:"Delete temporary FBX files after baking";
	button btnOpenFolder "Open Folder" width:140 tooltip: "Open folder with baked files";
	button btnBake "Bake" width:140 toolTip:"Bake selected maps in xNormal";
	
	struct mapStruct (name, desc, param, enabled);
	local maps = #(
		(mapStruct name:"base" desc:"Base Color Map" param:"%gXNGenBaseTex" enabled:"false"),
		(mapStruct name:"normal" desc:"Normal Map" param:"%gXNGenNormals" enabled:"true"),
		(mapStruct name:"height" desc:"Height Map" param:"%gXNGenHeights" enabled:"false"),
		(mapStruct name:"cavity" desc:"Cavity Map" param:"%gXNGenCavity" enabled:"false"),
		(mapStruct name:"convexity" desc:"Convexity Map" param:"%gXNGenConvexity" enabled:"false"),
		(mapStruct name:"thickness" desc:"Thickness Map" param:"%gXNGenThickness" enabled:"false"),
		(mapStruct name:"occlusion" desc:"Ambient Occlusion Map" param:"%gXNGenAO" enabled:"false"),
		(mapStruct name:"bent" desc:"Bent Normals Map" param:"%gXNGenBent" enabled:"false"),
		(mapStruct name:"vcols" desc:"Vertex Color Map" param:"%gXNBakeHighpolyVCols" enabled:"false"));
		
	local scriptPath = getDir #userStartupScripts + "\\Candygun\\";
	local defaultConfigFileName = scriptPath + "xnormal_default.xml";
	local configFileName = scriptPath + "xnormal_settings.xml";
	local appPath = "C:\\Program Files\\Santiago Orgaz\\xNormal 3.18.8\\x64\\";
	local appExe = "xNormal.exe";
	local lowPolyMesh;
	--local highPolyMesh;
	local baking = false;
	
	local settings = #();
	
	fn selectDropdownList list value = (
		list.selection = findItem list.items value;
	)
	
	fn exportObjectsFBX objs fileName highpoly:false cage:false =
	(
		with redraw off
		(
			local FBX_path = maxFilePath; --@"C:\FBX\";
			--makeDir FBX_path all:true;
			
			FbxExporterSetParam "Animation" false;
			FbxExporterSetParam "ASCII" false;
			FbxExporterSetParam "AxisConversionMethod" "None";
			FbxExporterSetParam "Cameras" false;
			FbxExporterSetParam "CAT2HIK" false;
			FbxExporterSetParam "EmbedTextures" false;
			FbxExporterSetParam "FileVersion" "FBX201300";
			FbxExporterSetParam "GeomAsBone" false;
			FbxExporterSetParam "GenerateLog" false;
			FbxExporterSetParam "Lights" false;
			FbxExporterSetParam "NormalsPerPoly" false;
			FbxExporterSetParam "PointCache" false;
			FbxExporterSetParam "Preserveinstances" false;
			FbxExporterSetParam "ScaleFactor" 1;
			FbxExporterSetParam "SelectionSetExport" false;
			FbxExporterSetParam "Shape" false;
			FbxExporterSetParam "Skin" false;
			FbxExporterSetParam "ShowWarnings" true;
			FbxExporterSetParam "SmoothingGroups" true; --(not (highpoly or cage));
			FbxExporterSetParam "SmoothMeshExport" false;
			FbxExporterSetParam "TangentSpaceExport" true; --false; --(not (highpoly or cage));
			FbxExporterSetParam "Triangulate" true;
			FbxExporterSetParam "UpAxis" "Z";
			FbxExporterSetParam "PreserveEdgeOrientation" true;

	
			local original_selection = getCurrentSelection();
			
			--local original_pos = obj.pos;
			--obj.pos = [0, 0, 0];
			select (for o in objs collect o);
			--selectmore o.children;

			local triangulate;
			if highpoly == false then (
				triangulate = Turn_To_Poly limitPolySize:on maxPolySize:3 name:"triangulate";
				addModifier $ triangulate;
			)

			exportFile fileName #noPrompt selectedOnly:true;
			
			if highpoly == false then
				deleteModifier $ triangulate;
			
			print ("exported " + fileName); 
			--obj.pos = original_pos;
			select original_selection;
		)
		return fileName;
	)	

	fn getSetting id = (
		for s in settings do (
			if s[1] == id then
				return s[2];
		)
		return "";
	)
	
	fn setSetting id value = (
		setUserProp lowPolyMesh id value;
		for s in settings do (
			if s[1] == id then (
				s[2] = value;
				return true;
			)
		)
		append settings #(id, value);
	)
	
	fn getSettings = (
		local name = "";
		if lowPolymesh != undefined then
			name = lowPolyMesh.name;
		
		-- Default values:		
		append settings #("%gXNHighPolyMesh", "");
		append settings #("%gXNHighBaseTex", "");
		append settings #("%gXNLowPolyMesh", "");
		append settings #("%gXNUseCage", "true");
		append settings #("%gXNCageFile", "");
		append settings #("%gXNBitmapX", "1024");
		append settings #("%gXNBitmapY", "1024");
		append settings #("%gXNEdgePadding", "16");
		append settings #("%gXNTexture", "T_" + name);
		append settings #("%gXNTangentSpace", "true");
		append settings #("%gXNDeleteExportedModels", "true");
		append settings #("%gXNAA", "2");
		
		for map in maps do
			append settings #(map.param, map.enabled);
		

		-- get stored settings from low poly mesh:
		if lowPolymesh != undefined then (
			for c in 1 to settings.count do (
				local value = getUserProp lowPolyMesh settings[c][1];
				if value != undefined then (
					settings[c][2] = value as String;
				)
			)
		)
		
		local fileFormat = trimLeft (getFileNameType (getSetting "%gXNTexture")) ".";
		if fileFormat == "" then
			fileFormat = "tga";
		
		selectDropdownList ddlSize (getSetting "%gXNBitmapX");
		selectDropdownList ddlFormat fileFormat;
		selectDropdownList ddlAntialiasing ("x" + (getSetting "%gXNAA"));
		selectDropdownList ddlPadding (getSetting "%gXNEdgePadding");
		chkBaseColor.checked = getSetting("%gXNGenBaseTex") == "true";
		chkNormal.checked = getSetting("%gXNGenNormals") == "true";
		chkOSNormal.checked = getSetting("%gXNTangentSpace") == "false";
		chkAO.checked = (getSetting "%gXNGenAO") == "true";
		chkCavity.checked = (getSetting "%gXNGenCavity") == "true";
		chkConvexity.checked = (getSetting "%gXNGenConvexity") == "true";
		chkVC.checked = (getSetting "%gXNBakeHighpolyVCols") == "true";
		chkThickness.checked = (getSetting "%gXNGenThickness") == "true";
		chkUseCage.checked = (getSetting "%gXNUseCage") == "true";
		chkDeleteExportedModels.checked = (getSetting "%gXNDeleteExportedModels") == "true";
	)

	fn storeSettings = (
		
		setSetting "%gXNBitmapX" ddlSize.selected;
		setSetting "%gXNBitmapY" ddlSize.selected;
		setSetting "%gXNAA" ddlAntialiasing.selected[2];
		setSetting "%gXNEdgePadding" ddlPadding.selected;
		setSetting "%gXNGenBaseTex" (chkBaseColor.checked as string);
		setSetting "%gXNGenNormals" ((chkNormal.checked or chkOSNormal.checked) as string);
		setSetting "%gXNTangentSpace" ((not chkOSNormal.checked) as string);
		setSetting "%gXNGenAO" (chkAO.checked as string);
		setSetting "%gXNGenCavity" (chkCavity.checked as string);
		setSetting "%gXNGenConvexity" (chkConvexity.checked as string);
		setSetting "%gXNBakeHighpolyVCols" (chkVC.checked as string);
		setSetting "%gXNGenThickness" (chkThickness.checked as string);
		setSetting "%gXNUseCage" (chkUseCage.checked as string);
		setSetting "%gXNDeleteExportedModels" (chkDeleteExportedModels.checked as string);
		setSetting "%gXNTexture" (maxFilePath + "T_" + lowPolymesh.name + "." + ddlFormat.selected);

		local projectionModifier = lowPolyMesh.modifiers[#projection]
		local lowPolyFile = (exportObjectsFBX #(lowPolyMesh) (maxfilepath + lowPolyMesh.name + ".fbx"));
		local highPolyFile = (exportObjectsFBX projectionModifier.geomNodes (maxfilepath + lowPolyMesh.name + "_highpoly.fbx") highpoly:true);
		local cageFile = "";

		local highPolyModel = projectionModifier.geomNodes[1];
		local highBaseTex = "";
		if highPolyModel != undefined and highpolyModel.mat != undefined and classOf highpolyModel.mat == standard and highpolyModel.mat.diffuseMap != undefined then
			highBaseTex = highpolyModel.mat.diffuseMap.filename;
		
		setSetting "%gXNHighBaseTex" highBaseTex;
		
		-- export cage:
		if (getSetting "%gXNUseCage") == "true" then (
			local oldSelection = getCurrentSelection();
			select lowPolymesh;
			local cageName = lowPolyMesh.name + "_cage";
			projectionModifier.exportCage cageName;
			local cage = getNodeByName cageName;
			if cage != undefined then (
				cageFile = (exportObjectsFBX #(cage) (maxfilepath + cageName + ".fbx") cage:true);
				delete cage;
			)
			select oldSelection;
		)
		
		setSetting "%gXNHighPolyMesh" highPolyFile;
		setSetting "%gXNLowPolyMesh" lowPolyFile;
		setSetting "%gXNCageFile" cageFile;
	)

	fn readConfigFile fileName = (
		local file = openFile fileName;
		seek file #eof
		local maxlen = filepos file
		seek file 0
		local stream = readChars file maxlen errorAtEOF:false;
		close file;
		return stream;
	)

	fn writeConfigFile fileName stream = (
		local file = openFile fileName mode:"w"
		format "%" stream to:file;
		close file;
	)
	
	fn replaceParams &stream params = (
		for param in params do (
			local pos = findString stream param[1];
			if pos != undefined then (
				stream = replace stream pos param[1].count param[2];
			)
		)
	)
	
	fn getMapFilename oldSuffix old:true withPath:false = (
		local baseName = getSetting "%gXNTexture";
		local fileName = getFilenameFile baseName;
		local ext = getFilenameType baseName;
		local path = "";
		if withPath then
			path = getFilenamePath baseName;
		
		if old then
			return path + fileName + oldSuffix + ext
		else (
			local newSuffix = "";
		
			if oldSuffix == "_baseTexBaked" then newSuffix = "";
			if oldSuffix == "_normals" then (
				if getSetting "%gXNTangentSpace" == "true" then
					newSuffix = "_N"
				else
					newSuffix = "_OSN";
			)
			if oldSuffix == "_occlusion" then newSuffix = "_AO";
			
			return path + fileName + newSuffix + ext;
		)
	)

	fn renameOutputFile oldSuffix = (
		local oldPath = getMapFilename oldSuffix withPath:true;
		local newPath = getMapFilename oldSuffix withPath:true old:false;

		if doesFileExist oldPath then (
			if doesFileExist newPath then
				deleteFile newPath;
			renameFile oldPath newPath;
		)
	)

	fn renameFiles = (
		if getSetting "%gXNGenBaseTex" == "true" then
			renameOutputFile "_baseTexBaked";
		if getSetting "%gXNGenNormals" == "true" then
			renameOutputFile "_normals";
		if getSetting "%gXNGenAO" == "true" then
			renameOutputFile "_occlusion";
		if getSetting "%gXNBakeHighpolyVCols" == "true" then
			renameOutputFile "_vcols";
	)

	fn deleteExportedModels = (
		deleteFile (getSetting "%gXNHighPolyMesh");
		deleteFile (getSetting "%gXNLowPolyMesh");
		deleteFile (getSetting "%gXNCageFile");
	)
	
	fn launchXNormal params = (
		local stream = readConfigFile defaultConfigFileName;
		replaceParams &stream params;
		writeConfigFile configFileName stream;
		
		local launchParam = appExe + " " + configFileName;
		hiddenDOSCommand launchParam startPath:appPath;
		print ("Started with : " + launchparam);
		renameFiles();
		if (getSetting "%gXNDeleteExportedModels") == "true" then
			deleteExportedModels();
	)

	
	fn bakeXNormal highPolymesh lowPolyMesh = (
		baking = true;
		setWaitCursor();
		print "storing settings...";
		storeSettings();
		launchXNormal settings;
		setArrowCursor();
		baking = false;
	)
	
	fn enableItems enable = (
		for control in XNormalRollout.controls do
			control.enabled = enable;
	)
	
	fn updateSelection = (
		if baking then
			return false;
		
		highpolyMeshes = #();
		lowPolyMesh = undefined;
		--lblLowpoly.text = "";
		--lblHighpoly.text = "";
		local sel = getCurrentSelection();
		
		if sel.count > 0 then (
			lowPolyMesh = sel[1];
			--lblLowpoly.text = "LP : " + lowPolyMesh.name;
		)
		
		/*if sel.count > 1 then (
			highPolyMesh = sel[2];
			
			if lowPolymesh.mesh.faces.count > highPolyMesh.mesh.faces.count then
				swap lowPolyMesh HighPolyMesh;
			lblLowPoly.text = "LP : " + lowPolyMesh.name;
			lblHighPoly.text = "HP : " + highPolyMesh.name;
		)*/
		
		getSettings();
		enableItems (sel.count > 0);
		--btnBake.enabled = sel.count == 1;
	)

	on XNormalRollout open do (
		callbacks.addScript #selectionSetChanged "XNormalRollout.updateSelection()" id:#XNormalRollout;
		updateSelection();
	)
	
	on XNormalRollout close do (
		callbacks.removeScripts #selectionSetChanged id:#XNormalRollout;
		xNormalRolloutOpen = false;
		updateToolbarButtons(); -- force button to display unchecked state
	)

	on btnOpenFolder pressed do (
		ShellLaunch (maxFilePath) "";
	)
	
	on btnAssignMaps pressed do (
		if lowPolyMesh == undefined then return();
		print "assigning maps...";
		
		local mat = lowPolyMesh.material;
		if mat == undefined then
			mat = Standard name:("M_" + lowPolyMesh.name);
		lowPolyMesh.material = mat;
		
		local normalMapFile = getMapFilename "_normals" withPath:true old:false;
		local hasNormalMap = (getSetting "%gXNGenNormals") == "true" and doesFileExist normalMapFile;
		mat.bumpMapEnable = hasNormalMap;
		mat.bumpMapAmount = 100;
		if hasNormalMap then (
			if mat.bumpMap == undefined then
				mat.bumpMap = normal_bump name:(lowPolyMesh.name + "_normal_bump");
			if mat.bumpMap.normal_map == undefined then
				mat.bumpMap.normal_map = bitmapTexture name:(fileNameFromPath normalMapFile);
			mat.bumpMap.normal_map.filename = fileNameFromPath normalMapFile;
		) else
			mat.bumpMap = undefined;
		
		local AOMapFile = getMapFilename "_occlusion" withPath:true old:false;
		local hasAOMap = (getSetting "%gXNGenAO") == "true" and doesFileExist AOMapFile;
		mat.diffuseMapEnable = hasAOMap;
		if hasAOMap then (
			if mat.diffuseMap == undefined then
				mat.diffuseMap = bitmapTexture name:(fileNameFromPath AOMapFile);
			mat.diffuseMap.filename = fileNameFromPath AOMapFile;
		) else
			mat.diffuseMap = undefined;
		
		showTextureMap mat on;
	)

	on btnBake pressed do (
		if doesFileExist (appPath + appExe) == false then (
			messageBox ("'" + appPath + appExe + "' not found. Please install xNormal.");
			return false;
		)
		
		if maxFilePath == "" then (
			messageBox "Please save your file before you bake maps.";
			return false;
		)
		
		if lowPolyMesh.modifiers[#projection] == undefined or lowPolyMesh.modifiers[#projection].geomNodes.count == 0 then (
			messageBox "Please assign a projection modifier to the low poly mesh and add some high poly meshes to it.";
			return false;
		)
		
		bakeXNormal highPolyMesh lowPolymesh;
	)
)

macroScript bake_xnormal category:"Candygun" buttonText: "xnormal" tooltip: "xnormal" autoUndoEnabled: false
(
	--on isEnabled do selection.count > 1;
	on isChecked do (
		return xNormalRolloutOpen;
	)
		
	on execute do (
		if (xNormalRolloutOpen) then (
			DestroyDialog XNormalRollout;
			xNormalRolloutOpen = false;
		) else (
			CreateDialog XNormalRollout;
			xNormalRolloutOpen = true;
		)
	)
)